/* Copyright (c) 2021, Haakan T. Johansson */
/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the authors nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
  gcc -O3 pattern.c -o pattern
 */

/* Program to search for serial pattern.
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

/* patterns
 *
 *   0    pp
 *   1   pp1
 *   2   p1p
 *   3   1pp
 *   4   pp0
 *   5   p0p
 *   6   0pp
 *
 *   7   ppv
 *   8   pvp
 *   9   vpp
 *
 * The below are not used, as we like to have a new bit every three cycles.
 *
 *   7   pp10
 *   8   pp01
 *
 *   9   p1p0
 * a 10  p0p1
 *
 * b 11  p10p
 * c 12  p01p
 *
 * d 13  1pp0
 * e 14  0pp1
 *
 * f 15  1p0p
 * g 16  0p1p
 *
 * h 17  10pp
 * i 18  01pp
 */

/*

47637431

aA0bBv0cC1dDeEvfF01gGhH1
..0..v0..1....v..01....1
aA.bB..cC.dDeE.fF..gGhH.

*/


/*
 * period: 28  vals: 7  (0)  dist 5:
 *
 * 2d6f83e0c /7  cgfgfbb  p01p 0p1p 1p0p 0p1p 1p0p p10p p10p
 *
 * 21f283e0f /7  fgfgif8  1p0p 0p1p 1p0p 0p1p 01pp 1p0p pp01
 */

int minreqdiff = 2;

uint64_t checked = 0;

uint64_t sequence;
int      seq_len;

#define NUM_AUX  1  /* 2 to search for patterns with 2 aux bits */

#if 1
#define SEQ_BITS     4
#define SEQ_MASK  0x0f
#define SEQ_MAX      9

/*                    0   1         4         7         10         15 */
int ptn_len[]     = { 2,  3, 3, 3,  3, 3, 3,  3, 3, 3,  0,0,0, 0,0,0, };
int ptn_vals[]    = { 1,  1, 1, 1,  1, 1, 1,  2, 2, 2,  0,0,0, 0,0,0, };
int ptn_mask1[]   = { 0,  1, 2, 4,  0, 0, 0,  0, 0, 0,  0,0,0, 0,0,0, };
int ptn_mask0[]   = { 0,  0, 0, 0,  1, 2, 4,  0, 0, 0,  0,0,0, 0,0,0, };
int ptn_maskv1p[] = { 1,  2, 1, 1,  2, 1, 1,  2, 1, 1,  0,0,0, 0,0,0, };
int ptn_maskv1n[] = { 2,  4, 4, 2,  4, 4, 2,  4, 4, 2,  0,0,0, 0,0,0, };
int ptn_maskv2[]  = { 0,  0, 0, 0,  0, 0, 0,  1, 2, 4,  0,0,0, 0,0,0, };
#endif

#if 0
#define SEQ_BITS     5
#define SEQ_MASK  0x1f
/*                    0   1         4         7      9      11     13     15     17     19     22         27      31 */     
int ptn_len[]     = { 2,  3, 3, 3,  3, 3, 3,  4, 4,  4, 4,  4, 4,  4, 4,  4, 4,  4, 4,  0,0,0, 0,0,0,0,0, 0,0,0,0,0, };
int ptn_vals[]    = { 1,  1, 1, 1,  1, 1, 1,  1, 1,  1, 1,  1, 1,  1, 1,  1, 1,  1, 1,  0,0,0, 0,0,0,0,0, 0,0,0,0,0, };
int ptn_mask1[]   = { 0,  1, 2, 4,  0, 0, 0,  2, 1,  4, 1,  4, 2,  8, 1,  8, 2,  8, 4,  0,0,0, 0,0,0,0,0, 0,0,0,0,0, };
int ptn_mask0[]   = { 0,  0, 0, 0,  1, 2, 4,  1, 2,  1, 4,  2, 4,  1, 8,  2, 8,  4, 8,  0,0,0, 0,0,0,0,0, 0,0,0,0,0, };
int ptn_maskv1p[] = { 1,  2, 1, 1,  2, 1, 1,  4, 4,  2, 2,  1, 1,  2, 2,  1, 1,  1, 1,  0,0,0, 0,0,0,0,0, 0,0,0,0,0, };
int ptn_maskv1n[] = { 2,  4, 4, 2,  4, 4, 2,  8, 8,  8, 8,  8, 8,  4, 4,  4, 4,  2, 2,  0,0,0, 0,0,0,0,0, 0,0,0,0,0, };
int ptn_maskv2[]  = { 0,  0, 0, 0,  0, 0, 0,  0, 0,  0, 0,  0, 0,  0, 0,  0, 0,  0, 0,  0,0,0, 0,0,0,0,0, 0,0,0,0,0, };
#endif

uint64_t mask_v2_a, mask_v2_b;
#if NUM_AUX == 2
uint64_t mask_v3_a, mask_v3_b;
#endif
uint64_t mask_0, mask_1;

uint64_t make_pattern(uint32_t value)
{
  uint64_t pattern = 0;
  uint64_t seq = sequence;
  int i;

  int x = 0;

  uint32_t v2 = value & 1;  value >>= 1;
#if NUM_AUX == 2
  uint32_t v3 = value & 1;  value >>= 1;
#endif
  uint32_t v = value;

  /* Fill out the bits of the pattern, using the fixed 0/1,
   * the value and the vars.
   */

  for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
    {
      uint32_t grp = seq & SEQ_MASK;
      int len  = ptn_len[grp];
      int vals = ptn_vals[grp];

      uint64_t v1 = (v     ) & 1;
      //uint64_t v2 = (v >> 1) & 1;

      uint64_t ptn_this =
	ptn_mask1[grp] |
	(ptn_maskv1p[grp] *  v1) |
	(ptn_maskv1n[grp] * !v1);// |
      //(ptn_maskv2[ptn] * v2);

      pattern |= ptn_this << x;

      x  += len;
      /* v >>= vals; */
      v >>= 1;
    }

  /* The location of the vars has been precalculated. */
  pattern |= (mask_v2_a * v2) | (mask_v2_b * !v2);

#if NUM_AUX == 2
  pattern |= (mask_v3_a * v3) | (mask_v3_b * !v3);
#endif
  
  return pattern;
}

uint32_t match_fixed(uint64_t pattern)
{
  uint64_t seq = sequence;
  int i;

  int diff = 0;

  /* Are the vars bit locations matching each other? */
  diff += (!(pattern & mask_v2_a)) == (!(pattern & mask_v2_b));
#if NUM_AUX == 2
  diff += (!(pattern & mask_v3_a)) == (!(pattern & mask_v3_b));
#endif

  for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
    {
      uint32_t grp = seq & SEQ_MASK;
      int len  = ptn_len[grp];
      int vals = ptn_vals[grp];

      int m1 = ptn_mask1[grp];
      int m0 = ptn_mask0[grp];
      int mvp = ptn_maskv1p[grp];
      int mvn = ptn_maskv1n[grp];

      /* Are the fixed bit locations matching? */
      diff += ((pattern & m1) != m1);
      diff += ((pattern & m0) != 0);

      /* Is the payload value bits matching? */
      diff += (!((pattern & mvp)) == (!(pattern & mvn)));
	
      pattern >>= len;
    }
  
  return diff;
}

int seq_period(int *period, int *values, int *vars)
{
  int i;
  uint64_t seq = sequence;

  /* Calculate the number of bit slots used by the sequence,
   * and the number of value and var bits, as well as
   * number of fixed 1 and 0.
   */
  
  *period = 0;
  *values = 0; /* Note: counts for each pair of bit-slots. */
  *vars = 0;   /* Note: counts for each bit-slot. */

  mask_0 = 0;
  mask_1 = 0;
  
  int n0 = 0, n1 = 0;

  for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
    {
      uint32_t grp = seq & SEQ_MASK;

      int len = ptn_len[grp];
      if (!len)
	return 0;

      //mask_0 |= ptn_mask0[grp] << *period;
      //mask_1 |= ptn_mask1[grp] << *period;

      n0 += !!ptn_mask0[grp];
      n1 += !!ptn_mask1[grp];
      *vars += !!ptn_maskv2[grp];
      if (!!ptn_maskv2[grp])
	{
	  /* Keep track of the two var bit locations. */
	  /* Since we will have 2, just keep two last values seen. */
#if NUM_AUX == 2
          mask_v3_a = mask_v3_b;
          mask_v3_b = mask_v2_a;
#endif
          mask_v2_a = mask_v2_b;
	  mask_v2_b = ptn_maskv2[grp] << *period;
	}
      *period += len;

      /* Every group has a value. */
      *values += 1; // ptn_vals[grp];
    }

  if (n0 != n1) // unbalanced
    return 0;

  if (*vars != NUM_AUX * 2)
    return 0;

  if (*period > 32)
    return 0;
  
  return 1;
}

int try_sequence(int *period, int *values, int *vars,
		 int *hasfour, int *hasfive)
{
  uint32_t value;
  int offset;
  int i;

  /* Check if any rotation of the sequence has a smaller value,
   * if so, we are not the canonical, so do no test.
   * (A rotation is effectively the same pattern.)
   *
   * Update 2021-01-12: NO, it is NOT!
   *
   * The pattern generated is the same, but the matching function
   * does not behave the same...
   */
  for (i = 1; i < seq_len; i++)
    {
      uint64_t seq2 =
	(sequence >> (i * SEQ_BITS)) |
	(sequence << ((seq_len - i) * SEQ_BITS));
      seq2 &= (((uint64_t) 1) << (seq_len*SEQ_BITS)) - 1;
      // printf ("%0.*llx  %0.*llx\n", seq_len, sequence, seq_len, seq2);
      if (seq2 < sequence)
	; // return -1; // see comment above, not-the-same!
    }

  /* Figure out number of values and vars.
   * Also refuse sequence if it has some simple bad properties.
   */
  if (!seq_period(period, values, vars))
    return -1;

  checked++;

  int overall_mindiff = 100;

  /* Total number of payload bits. */
  int val_plus_vars = *values+*vars/2;

  *hasfour = 0;
  *hasfive = 0;

  /* When generating patterns to test against, we need two
   * full sequences, so twice the number of payload bits.
   */
  uint32_t max_value = (1 << (2*(val_plus_vars)));

  uint64_t mask_check = ((((uint64_t) 1) << (*period)) -1);

  // printf ("val_plus_vars: %d, period: %d\n", val_plus_vars, *period);

  /* Check every possible payload in the two sequences. */
  // for (value = 0x12700; value <= 0x12700; value++)
  for (value = 0; value < max_value; value++)
    {
      /* Set up the actual data. */
      uint64_t pattern =
	make_pattern(value) |
	(make_pattern(value >> val_plus_vars)) << (*period);

      /*
      printf ("%04x: %016llx : ", value, pattern);

      for (i = 2*(*period)-1; i >= (*period); i--)
	printf ("%d", (pattern >> i) & 1);
      printf (" ");
      for (i = (*period)-1; i >= 0; i--)
	printf ("%d", (pattern >> i) & 1);
      printf (" : ");
      */

      int mindiff = 100;

      /* Calculate the difference at all offsets, except 0. */
      for (offset = 1; offset < (*period); offset++)
	{
	  uint64_t check = (pattern >> offset) & mask_check;
	  
	  int diff = match_fixed(check);

	  // printf (" %d%c", diff, diff == 0 ? '*' : ' ');

	  if (diff < mindiff /*&& offset != 0*/)
	    mindiff = diff;
	}

      /* Invert the test pattern. */
      pattern = ~pattern;

      /* Calculate the difference at all offsets, including 0. */ 
      for (offset = 0; offset < (*period); offset++)
	{
	  uint64_t check = (pattern >> offset) & mask_check;
	  
	  int diff = match_fixed(check);

	  // printf (" %d%c", diff, diff == 0 ? '*' : ' ');

	  if (diff < mindiff /*&& offset != 0*/)
	    mindiff = diff;
	}
      // printf (" : %d %c\n", mindiff, mindiff == 0 ? 'X' : ' ');

      /* Smaller difference than before? */
      if (mindiff < overall_mindiff)
	{
	  overall_mindiff = mindiff;

	  /* So small difference that we are not interested? */
	  if (overall_mindiff < minreqdiff)
	    return -1;
	}
    }

  /* Calculate if we may have four consequtive same bits only if we
   * survived this far.  This is only for information.
   */
  for (value = 0; value < max_value; value++)
    {
      uint64_t pattern =
	make_pattern(value) |
	(make_pattern(value >> val_plus_vars)) << (*period);

      for (offset = 0; offset < (*period); offset++)
	{
	  if (((pattern >> offset) & 0x0f) == 0x0f ||
	      ((pattern >> offset) & 0x0f) == 0x00)
	    *hasfour = 1;
	  if (((pattern >> offset) & 0x1f) == 0x1f ||
	      ((pattern >> offset) & 0x1f) == 0x00)
	    *hasfive = 1;
	}

      /* This is a debug check, no need to do unless we claim to be
       * good.
       */

      /* Check that we at offset 0 do not have a mismatch. */
      int diff = match_fixed(pattern);

      if (diff)
	{
	  fprintf (stderr, "Mismatch at offset 0!\n");
	  exit(1);
	}
    }
  
  return overall_mindiff;
}



int main(int argc, char *argv[])
{
  int period;
  int values;
  int vars;

  /* Stepping pattern for sequences. */
  int tot  = 1;
  int this = 0;

  int attempts = 0;

  if (argc >= 3)
    {
      this = strtol(argv[1],NULL,0);
      tot  = strtol(argv[2],NULL,0);
    }

  /* Which sequence lengths to try. */
  for (seq_len = 8; seq_len <= 8; seq_len++)
    {
      uint64_t max_seq = ((uint64_t) 1) << (seq_len*SEQ_BITS);

      /* Check all sequences. */
      for (sequence = this; sequence < max_seq; sequence += tot)
	{
	  // sequence = 0x13473674;
	  
	  uint64_t seq = sequence;
	  int i;

	  for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
	    {
	      uint64_t grp = seq & SEQ_MASK;

	      /* Is the group outside number of groups. */
	      if (grp > SEQ_MAX)
		{
		  // How much need to be added at minimum to reach
		  // the next valid number?

		  /* Next valid is if this and all lower groups
		   * start again at zero,
		   * while next group is incremented by 1.
		   */
		  uint64_t next_valid =
		    (sequence &
		     ~((((uint64_t) 1) << (SEQ_BITS * (i+1))) - 1)) +
		    (((uint64_t) 1) << (SEQ_BITS * (i+1)));
		  /*
		  printf ("%0.*llx -> %0.*llx (+%0.*llx)\n",
			  seq_len, sequence,
			  seq_len, next_valid,
			  seq_len, next_valid - sequence);
		  */
		  uint64_t min_add = next_valid - sequence;
		  if (min_add > tot)
		    {
		      sequence += ((min_add-1) / tot) * tot;
		    }
		  // Note, another tot will also be added.
		  goto try_next;
		}
	    }
	  /*
	  printf ("%0.*llx /%d\n", seq_len, sequence, seq_len);
	  fflush(stdout);
	  */  

	  /* We want to have a pattern that starts with a fixed
	   * bit.
	   */
	  {
	    uint64_t grp = sequence & SEQ_MASK;

	    if (!(ptn_mask0[grp] == 1) &&
		!(ptn_mask1[grp] == 1))
	      continue;
	  }

	  /* Find how long stretch of surplus 1s we have... */
	  /* In order to handle the periodicity, we run through
	   * the code twice.
	   */
	  int ones = 0;
	  int zeros = 0;
	  int maxones = 0;
	  int maxzeros = 0;
	  seq = sequence;
	  for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
	    {
	      uint64_t grp = seq & SEQ_MASK;

	      if (ptn_mask0[grp])
		ones = 0;
	      else
		{
		  ones++;
		  if (ones > maxones)
		    maxones = ones;
		}
	      if (ptn_mask1[grp])
		zeros = 0;
	      else
		{
		  zeros++;
		  if (zeros > maxzeros)
		    maxzeros = zeros;
		}
	      // printf ("ptn: %d %d\n",ones,zeros);
	    }
	  seq = sequence;
	  for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
	    {
	      uint64_t grp = seq & SEQ_MASK;

	      if (ptn_mask0[grp])
		ones = 0;
	      else
		{
		  ones++;
		  if (ones > maxones)
		    maxones = ones;
		}
	      if (ptn_mask1[grp])
		zeros = 0;
	      else
		{
		  zeros++;
		  if (zeros > maxzeros)
		    maxzeros = zeros;
		}
	      // printf ("ptn: %d %d\n",ones,zeros);
	    }
	  /* Avoid patterns that give more than 3 extra 0 or 1 in a row.
	   */
	  if (maxones > 3)
	    continue;
	  if (maxzeros > 3)
	    continue;

	  attempts++;
	  if (!(attempts & 0xfff))
	    {
	      printf ("%0.*llx /%d\r", seq_len, sequence, seq_len);
	      fflush(stdout);
	    }

	  int hasfour;
	  int hasfive;
	  int dist = try_sequence(&period, &values, &vars, &hasfour, &hasfive);

	  if (dist < 0)
	    continue;

	  /* printf ("%0.*llx /%d\r", seq_len, sequence, seq_len); */

	  if (dist > 1)
	    {
	      printf ("%0.*llx /%d", seq_len, sequence, seq_len);
	      printf ("   ");
	      {
		int i;
		uint64_t seq = sequence;
		for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
		  {
		    uint32_t ptn = seq & SEQ_MASK;
		    printf ("%c", ptn < 10 ? ptn + '0' : ptn - 10 + 'a');
		  }
	      }
	      printf ("   ");	      
	      printf ("period: %2d  vals: %d  (%d)  ", period, values, vars);
	      printf ("dist %d four %d five %d  ", dist, hasfour, hasfive);
	      {
		int i;
		uint64_t seq = sequence;
		int nv2 = 0;
		for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
		  {
                    uint32_t ptn = seq & SEQ_MASK;
		    int j;
		    for (j = 0; j < ptn_len[ptn]; j++)
		      {
			if (ptn_mask0[ptn] & (1 << j))
			  printf ("0");
			if (ptn_mask1[ptn] & (1 << j))
			  printf ("1");
			if (ptn_maskv1p[ptn] & (1 << j))
			  printf ("%c", 'a' + i);
			if (ptn_maskv1n[ptn] & (1 << j))
			  printf ("%c", 'A' + i);
			if (ptn_maskv2[ptn] & (1 << j))
			  {
			    printf ("%c", nv2 ? 'V' : 'v');
			    nv2++;
			  }
		      }
		  }
		printf (" ");
		seq = sequence;
		for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
		  {
                    uint32_t ptn = seq & SEQ_MASK;
		    int j;
		    for (j = 0; j < ptn_len[ptn]; j++)
		      {
			if (ptn_mask0[ptn] & (1 << j))
			  printf ("0");
			if (ptn_mask1[ptn] & (1 << j))
			  printf ("1");
			if (ptn_maskv1p[ptn] & (1 << j) ||
			    ptn_maskv1n[ptn] & (1 << j))
			  printf (".");
			if (ptn_maskv2[ptn] & (1 << j))
			  printf ("-");
		      }
		  }

		seq = sequence;
		int mask01v2[20];
		for (i = 0; i < seq_len; i++, seq >>= SEQ_BITS)
		  {
		    uint32_t ptn = seq & SEQ_MASK;
		    mask01v2[i] =
		      ptn_mask0[ptn] | ptn_mask1[ptn] | ptn_maskv2[ptn];
		  }
		printf (" %d%d %d%d %d%d",
			(mask01v2[0] == mask01v2[1] &&
			 mask01v2[0] == mask01v2[2] &&
			 mask01v2[0] == mask01v2[3]),
			(mask01v2[4] == mask01v2[5] &&
			 mask01v2[4] == mask01v2[6] &&
			 mask01v2[4] == mask01v2[7]),
			(mask01v2[0] == mask01v2[1] &&
			 mask01v2[0] == mask01v2[4] &&
			 mask01v2[0] == mask01v2[5]),
			(mask01v2[2] == mask01v2[3] &&
			 mask01v2[2] == mask01v2[6] &&
			 mask01v2[2] == mask01v2[7]),
			(mask01v2[0] == mask01v2[2] &&
			 mask01v2[0] == mask01v2[4] &&
			 mask01v2[0] == mask01v2[6]),
			(mask01v2[1] == mask01v2[3] &&
			 mask01v2[1] == mask01v2[5] &&
			 mask01v2[1] == mask01v2[7]));
	      }
	      printf ("\n");
	    }
	  //else
	  //  printf ("\r");

	try_next:
	  ;
	}
    }

  printf ("checked: %lld\n", checked);
}
