-- Copyright (c) 2021, Haakan T. Johansson
-- All rights reserved.

-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the authors nor the
--       names of its contributors may be used to endorse or promote products
--       derived from this software without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
-- AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
-- LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
-- SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
-- INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
-- CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
-- ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-- POSSIBILITY OF SUCH DAMAGE.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package rataser_records is

  -- This record is not used, since it caused various problems with
  -- broken tools (see rataser_sym24word8_util.vhd).
  -- type rataser_sym24def is record
  --   -- Slots for 8 bit values, positive and inverted
  --   sp : rataser_int_array(0 to 7);
  --   sn : rataser_int_array(0 to 7);
  --   --sn : array (0 to 7) of integer;
  --   -- Slots for 1 aux value, positive and inverted
  --   vp : integer;
  --   vn : integer;
  --   -- Slots for 3 fixed 0 and 1 values
  --   f0 : rataser_int_array(0 to 2);
  --   f1 : rataser_int_array(0 to 2);
  -- end record rataser_sym24def;

  type rataser_sync_count_in is record
    -- Markers for good/bad events.
    -- Both up and down = '1' makes no sense, but down will win.
    -- Good thing happened (affirming sync).
    up               : std_logic;
    -- Bad thing happened (challenging sync).
    down             : std_logic;
    -- Possible sync track start point found.
    tentative        : std_logic;
  end record;

  constant rataser_sync_count_in_zero : rataser_sync_count_in :=
    (up        => '0',
     down      => '0',
     tentative => '0');

  type rataser_sync_count_out is record
    -- When synchronised, or starting to try to track.
    tentative_synced : std_logic;
    -- When synchronised.
    synced           : std_logic;
  end record;

  constant rataser_sync_count_out_zero : rataser_sync_count_out :=
    (tentative_synced => '0',
     synced           => '0');

  type rataser_sync_count_iface is record
    i : rataser_sync_count_in;
    o : rataser_sync_count_out;
  end record;

  constant rataser_sync_count_iface_zero : rataser_sync_count_iface :=
    (i => rataser_sync_count_in_zero,
     o => rataser_sync_count_out_zero);

  type rataser_sym8word1_debug is record
    accum_min_max  : std_logic_vector(15 downto 0);
    accum_max_min  : std_logic_vector(15 downto 0);
    accum_avg      : std_logic_vector(15 downto 0);
    accum_this     : std_logic_vector(15 downto 0);
    accum_high_max : std_logic_vector(15 downto 0);
    accum_high_ind : std_logic_vector( 3 downto 0);
  end record;

end rataser_records;
